unit ALTTA_Switch;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, System.ImageList, Vcl.ImgList, Vcl.StdCtrls;


const
  AM_NOTIFY = WM_APP + WM_NOTIFY;

type
  TSwitchForm = class(TForm)
    ShellImages: TImageList;
    TaskList: TListBox;
    procedure ApplicationIdle(Sender: TObject; var Done: Boolean);
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure FormDeactivate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure TaskListDrawItem(Control: TWinControl; Index: Integer; Rect: TRect;
      State: TOwnerDrawState);
    procedure TaskListKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure TaskListClick(Sender: TObject);
  private
    { Private 錾 }
    FFlags: Cardinal;
    FMargin: Integer;
    procedure ReadOption;
    procedure WriteOption;
    function Altered: Boolean;
    procedure AlteringStart;
    procedure AlteringStop;
    procedure AlteringRestart;
    procedure ActivateInFront(Wnd: HWND);
    procedure MaskMenuKey;
    procedure SelectNextItem;
    procedure SelectPrevItem;
    procedure SelectCurrentItem;
    procedure WMHotKey(var Message: TWMHotKey); message WM_HOTKEY;
    procedure WMDisplayChange(var Message: TMessage); message WM_DISPLAYCHANGE;
    procedure AMNotify(var Message: TMessage); message AM_NOTIFY;
  protected
    { protected 錾 }
    procedure CreateParams(var Params: TCreateParams); override;
  public
    { Public 錾 }
    ListWidth,
    IconSize: Integer;
    Modifier: Word;
  end;

var
  SwitchForm: TSwitchForm;

implementation

{$R *.dfm}

uses
  ALTTA_Config,
  System.UITypes, Hotkeys, Winapi.Dwmapi, WndInfo, AppData, Winapi.ShellAPI, System.Math;

type
  tagKBDLLHOOKSTRUCT = record
    vkCode: DWORD;
    scanCode: DWORD;
    flags: DWORD;
    time: DWORD;
    dwExtraInfo: ULONG_PTR;
  end;
  TKbdLLHookStruct = tagKBDLLHOOKSTRUCT;
  PKbdLLHookStruct = ^TKbdLLHookStruct;

const
  LLKHF_EXTENDED = KF_EXTENDED shr 8;
  LLKHF_LOWER_IL_INJECTED = $00000002;
  LLKHF_INJECTED = $00000010;
  LLKHF_ALTDOWN = KF_ALTDOWN shr 8;
  LLKHF_UP = KF_UP shr 8;

  DWMWA_CLOAKED = 14;

  ForwardID = VK_MENU;
  BackwardID = VK_MENU + VK_SHIFT;

var
  KeyboardHook: HHOOK = 0;

  Modifiers: Word = 0;
  TriggerKEY,
  ForwardMOD,
  BackwardMOD: Word;

{ Routine }

function LowLevelKeyboardProc(nCode: Integer;
  wParam: WPARAM; lParam: PKbdLLHookStruct): LRESULT; stdcall;
begin
  Result := 0;

  if nCode = HC_ACTION then
  begin
    case wParam of
      WM_KEYDOWN, WM_SYSKEYDOWN:
      begin
        case lParam.vkCode of
          VK_LSHIFT, VK_RSHIFT:
            if (Modifiers and MOD_SHIFT) = 0 then Inc(Modifiers, MOD_SHIFT);
          VK_LCONTROL, VK_RCONTROL:
            if (Modifiers and MOD_CONTROL) = 0 then Inc(Modifiers, MOD_CONTROL);
          VK_LMENU, VK_RMENU:
            if (Modifiers and MOD_ALT) = 0 then Inc(Modifiers, MOD_ALT);
          VK_LWIN, VK_RWIN:
            if (Modifiers and MOD_WIN) = 0 then Inc(Modifiers, MOD_WIN);
        end;

        if (lParam.vkCode = TriggerKEY) and (Modifiers = ForwardMOD) then
        begin
          PostMessage(SwitchForm.Handle, WM_HOTKEY, ForwardID, MakeLong(Modifiers, TriggerKEY));
          Result := 1;
        end
        else if (lParam.vkCode = TriggerKEY) and (Modifiers = BackwardMOD) then
        begin
          PostMessage(SwitchForm.Handle, WM_HOTKEY, BackwardID, MakeLong(Modifiers, TriggerKEY));
          Result := 1;
        end;
      end;
      WM_KEYUP, WM_SYSKEYUP:
      begin
        case lParam.vkCode of
          VK_LSHIFT, VK_RSHIFT:
            if (Modifiers and MOD_SHIFT) <> 0 then Dec(Modifiers, MOD_SHIFT);
          VK_LCONTROL, VK_RCONTROL:
            if (Modifiers and MOD_CONTROL) <> 0 then Dec(Modifiers, MOD_CONTROL);
          VK_LMENU, VK_RMENU:
            if (Modifiers and MOD_ALT) <> 0 then Dec(Modifiers, MOD_ALT);
          VK_LWIN, VK_RWIN:
            if (Modifiers and MOD_WIN) <> 0 then Dec(Modifiers, MOD_WIN);
        end;
      end;
    end;
  end;

  if Result = 0 then
    Result := CallNextHookEx(KeyboardHook, nCode, wParam, Winapi.Windows.LPARAM(lParam));
end;

function IncludeThis(Wnd: HWND): Boolean;
var
  Cloaked: Cardinal;
begin
  Result := IsWindowVisible(Wnd) and ((GetWindowLong(Wnd, GWL_EXSTYLE) and WS_EX_TOOLWINDOW) = 0);
  if Result and TOSVersion.Check(6) then
  begin
    if Succeeded(DwmGetWindowAttribute(Wnd, DWMWA_CLOAKED, @Cloaked, SizeOf(Cardinal))) then
      Result := Cloaked = 0;
  end;
end;

function EnumWindowsProc(hwnd: HWND; lParam: LPARAM): BOOL; stdcall;
begin
  Result := True;

  if IncludeThis(hwnd) then
    SwitchForm.TaskList.Items.AddObject(GetWindowTextStr(hwnd), TObject(hwnd));
end;

{ TSwitchForm }

procedure TSwitchForm.ReadOption;
var
  IniFile: TAppIniFile;
begin
  IniFile := TAppIniFile.Create(AppDataName('ShaggyDog\Alttabalt\Alttabalt.ini'));
  try
    //View
    ListWidth := IniFile.ReadInteger('View', 'ListWidth', 480);
    IconSize := IniFile.ReadInteger('View', 'IconSize', SHGFI_LARGEICON);
    TaskList.Font.Name := IniFile.ReadString('View', 'FontName', 'MS UI Gothic');
    TaskList.Font.Size := IniFile.ReadInteger('View', 'FontSize', 12);
    TaskList.Font.Style := TFontStyles(Byte(IniFile.ReadInteger('View', 'FontStyle', 0)));
    //Hotkey
    Modifier := IniFile.ReadInteger('Hotkey', 'Modifier', MOD_ALT);
  finally
    IniFile.Free;
  end;
end;

procedure TSwitchForm.WriteOption;
var
  IniFile: TAppIniFile;
begin
  IniFile := TAppIniFile.Create(AppDataName('ShaggyDog\Alttabalt\Alttabalt.ini'));
  try
    //View
    IniFile.WriteInteger('View', 'ListWidth', ListWidth);
    IniFile.WriteInteger('View', 'IconSize', IconSize);
    IniFile.WriteString('View', 'FontName', TaskList.Font.Name);
    IniFile.WriteInteger('View', 'FontSize', TaskList.Font.Size);
    IniFile.WriteInteger('View', 'FontStyle', Byte(TaskList.Font.Style));
    //Hotkey
    IniFile.WriteInteger('Hotkey', 'Modifier', Modifier);
  finally
    IniFile.Free;
  end;
end;

function TSwitchForm.Altered: Boolean;
begin
  Result := KeyboardHook <> 0;
end;

procedure TSwitchForm.AlteringStart;
var
  FileInfo: TSHFileInfo;
begin
  if not Altered then
  begin
    Width := ListWidth;
    FFlags := SHGFI_SYSICONINDEX or IconSize;
    FMargin := 2;
    case IconSize of
      SHGFI_LARGEICON: ShellImages.Width := 32;
      SHGFI_SMALLICON: ShellImages.Width := 16;
    end;
    ShellImages.Height := ShellImages.Width;
    ShellImages.Handle := SHGetFileInfo('.', 0, FileInfo, SizeOf(FileInfo), FFlags);
    TaskList.Canvas.Font.Assign(TaskList.Font);
    TaskList.ItemHeight := Max(ShellImages.Height, TaskList.Canvas.TextHeight('Ag')) + FMargin * 2;

    TriggerKEY := VK_TAB;
    ForwardMOD := Modifier;
    BackwardMOD := Modifier + MOD_SHIFT;
    KeyboardHook := SetWindowsHookEx(WH_KEYBOARD_LL, @LowLevelKeyboardProc, hInstance, 0);
    if KeyboardHook = 0 then
      MessageDlg('WH_KEYBOARD_LL' + #13#10 + SysErrorMessage(GetLastError), mtError, [mbOk], 0);
  end;
end;

procedure TSwitchForm.AlteringStop;
begin
  if KeyboardHook <> 0 then
  begin
    UnhookWindowsHookEx(KeyboardHook);
    KeyboardHook := 0;
  end;
end;

procedure TSwitchForm.AlteringRestart;
begin
  AlteringStop;
  try
    Application.ProcessMessages;

    Position := poScreenCenter;
  finally
    AlteringStart;
  end;
end;

procedure TSwitchForm.ActivateInFront(Wnd: HWND);
var
  AltDown: Boolean;
begin
  if Wnd <> 0 then
  begin
    if IsIconic(Wnd) then
      ShowWindowAsync(Wnd, SW_RESTORE);

    //őOʂŃANeBuɂB
    //https://www.codeproject.com/Tips/76427/How-to-bring-window-to-top-with-SetForegroundWindo
    AltDown := GetKeyState(VK_MENU) < 0;
    if not AltDown then
      keybd_event(VK_MENU, 0, KEYEVENTF_EXTENDEDKEY, 0);
    SetForegroundWindow(Wnd);
    if not AltDown then
      keybd_event(VK_MENU, 0, KEYEVENTF_EXTENDEDKEY or KEYEVENTF_KEYUP, 0);
  end;
end;

procedure TSwitchForm.MaskMenuKey;
begin
  //Alt L[_E܂B
  //https://www.haijin-boys.com/discussions/4583
  if GetKeyState(VK_MENU) < 0 then
  begin
    keybd_event(VK_CONTROL, 0, KEYEVENTF_EXTENDEDKEY, 0);
    keybd_event(VK_CONTROL, 0, KEYEVENTF_EXTENDEDKEY or KEYEVENTF_KEYUP, 0);
  end;
end;

procedure TSwitchForm.SelectNextItem;

  procedure IncItemIndex(Index: Integer);
  begin
    if TaskList.Items.Count > 0 then
    begin
      Inc(Index);
      if Index >= TaskList.Items.Count then
        Index := 0;
      TaskList.Selected[Index] := True;
    end;
  end;

begin
  if Visible then
  begin
    SetForegroundWindow(Handle);
    IncItemIndex(TaskList.ItemIndex);
  end
  else
  begin
    TaskList.Items.Clear;
    EnumWindows(@EnumWindowsProc, 0);
    if TaskList.Items.Count > 0 then
    begin
      MaskMenuKey;
      IncItemIndex(0);
      Show;
      SetForegroundWindow(Handle);
    end;
  end;
end;

procedure TSwitchForm.SelectPrevItem;

  procedure DecItemIndex(Index: Integer);
  begin
    if TaskList.Items.Count > 0 then
    begin
      Dec(Index);
      if Index < 0 then
        Index := TaskList.Items.Count - 1;
      TaskList.Selected[Index] := True;
    end;
  end;

begin
  if Visible then
  begin
    SetForegroundWindow(Handle);
    DecItemIndex(TaskList.ItemIndex);
  end
  else
  begin
    TaskList.Items.Clear;
    EnumWindows(@EnumWindowsProc, 0);
    if TaskList.Items.Count > 0 then
    begin
      MaskMenuKey;
      DecItemIndex(TaskList.Items.Count);
      Show;
      SetForegroundWindow(Handle);
    end;
  end;
end;

procedure TSwitchForm.SelectCurrentItem;
var
  ModKey: Word;
begin
  if Altered then
  begin
    ModKey := GetModifyKeys;
    if (ForwardMOD <> ModKey) and (BackwardMOD <> ModKey) then
    begin
      try
        if TaskList.ItemIndex >= 0 then
          ActivateInFront(HWND(TaskList.Items.Objects[TaskList.ItemIndex]));
      finally
        Hide;
      end;
    end;
  end;
end;

procedure TSwitchForm.ApplicationIdle(Sender: TObject; var Done: Boolean);
begin
  //TaskList.OnKeyUp ȂƂ̂ŁAAChCxgł`FbNB
  SelectCurrentItem;
end;

procedure TSwitchForm.CreateParams(var Params: TCreateParams);
begin
  inherited;

  //{[_[gɂB
  Params.Style := Params.Style and not (WS_DLGFRAME or WS_THICKFRAME) or WS_POPUP;
  //^XNo[{^BApplication.MainFormOnTaskbar = True ł邱ƁB
  Params.ExStyle := Params.ExStyle and not WS_EX_APPWINDOW;
  Params.WndParent := Application.Handle;
end;

procedure TSwitchForm.FormCreate(Sender: TObject);
begin
  ReadOption;

  Application.ShowMainForm := False;
  Application.OnDeactivate := FormDeactivate;
  if TOSVersion.Check(6) then
    ChangeWindowMessageFilter(AM_NOTIFY, MSGFLT_ADD);
  AlteringStart;
end;

procedure TSwitchForm.WMHotKey(var Message: TWMHotKey);
var
  Wnd: HWND;
  WndName, WndTitle: string;
begin
  Wnd := GetForegroundWindow;
  WndName := GetClassNameStr(Wnd);
  WndTitle := GetWindowTextStr(Wnd);
  if (WndTitle <> 'Task Switching')
    and ((WndName <> 'TaskSwitcherWnd') or (WndName <> 'Task Switching')) then
  begin
    if Message.HotKey = ForwardID then
      SelectNextItem
    else if Message.HotKey = BackwardID then
      SelectPrevItem;
  end;
end;

procedure TSwitchForm.WMDisplayChange(var Message: TMessage);
begin
  Monitor;  //TScreen.Monitors ̕s΍B͎Ă݂ǁAꉞB

  AlteringRestart;
end;

procedure TSwitchForm.AMNotify(var Message: TMessage);
begin
  inherited;

  SetForegroundWindow(Handle);
  case ConfigDialog.ShowModal of
    mrOk:
    begin
      WriteOption;
      AlteringRestart;
    end;
    mrCancel: ;
    mrClose: Close;
  end;
end;

procedure TSwitchForm.FormShow(Sender: TObject);
begin
  ClientHeight := Min(
    (TaskList.ItemHeight * TaskList.Items.Count) + FMargin * 2,
    Monitor.WorkareaRect.Height - FMargin * 2
  );
  Application.OnIdle := ApplicationIdle;
end;

procedure TSwitchForm.FormHide(Sender: TObject);
begin
  Application.OnIdle := nil;
  Position := poScreenCenter;
end;

procedure TSwitchForm.FormDeactivate(Sender: TObject);
begin
  if SwitchForm.Visible then
    SwitchForm.Hide;
end;

procedure TSwitchForm.FormDestroy(Sender: TObject);
begin
  AlteringStop;
end;

procedure TSwitchForm.TaskListDrawItem(Control: TWinControl; Index: Integer; Rect: TRect;
  State: TOwnerDrawState);
var
  ListBox: TListBox;
  s: string;
  FileInfo: TSHFileInfo;
  TextTop, TextLeft, TextHeight: Integer;
begin
  ListBox := TListBox(Control);
  ListBox.Canvas.FillRect(Rect);
  s := GetWindowFileName(HWND(ListBox.Items.Objects[Index]));
  if Succeeded(SHGetFileInfo(PChar(s), 0, FileInfo, SizeOf(FileInfo), FFlags)) then
  begin
    ShellImages.Draw(ListBox.Canvas,
      Rect.Left + FMargin, Rect.Top + (Rect.Height - ShellImages.Height) div 2, FileInfo.iIcon);
  end;
  s := ListBox.Items.Strings[Index];
  TextHeight := ListBox.Canvas.TextHeight(s);
  TextLeft := Rect.Left + ShellImages.Width + FMargin * 3;
  TextTop := Rect.Top + (Rect.Height - TextHeight) div 2;
  ListBox.Canvas.TextOut(TextLeft, TextTop, s);
end;

procedure TSwitchForm.TaskListKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  //Ô߁AG^[ƃGXP[vB
  if Visible then
    case Key of
      VK_RETURN: SelectCurrentItem;
      VK_ESCAPE: Close;
    else
      SelectCurrentItem;
    end;
end;

procedure TSwitchForm.TaskListClick(Sender: TObject);
begin
  SelectCurrentItem;
end;

end.
